/*
    
                    The Hacker's Diet

              Trend Fitting
              
    This file contains code which fits a linear trend to a
    collection of day to day trend values using linear regression
    with the least squares method.  Unlike the majority of the
    application, this code does resort to floating point; maybe
    some hero will step forward to convert it to overflow-safe
    fixed point with adequate precision.
             
*/

#include <stdio.h>
#include <assert.h>

#include "database.h"
#include "memdb.h"
#include "units.h"
#include "trendfit.h"

#define TrendToDouble(t) ((t) / 1000000.0)

static int n = 0;
static double s1, s2, s3, s4, s5;   /* Terms and factors in linear regression relation */

/*  fitSlopeStart  --  Initialise trend fit  */

void fitSlopeStart(void)
{
    assert(n == 0);
    n = 0;
    s1 = s2 = s3 = s4 = 0;
}

/*  fitSlopePoint  --  Add a point on the trend curve to the calculation  */

void fitSlopePoint(trend t)
{
    double td = TrendToDouble(t);
    
    s1 += (n + 1) * td;
    s2 += (n + 1);
    s3 += td;
    s4 += ((double) (n + 1)) * (n + 1);
    n++;
}

/*  fitSlopeEnd  --  Complete calculation and return best fit slope  */

double fitSlopeEnd(void)
{
    s1 *= n;
    s4 *= n;
    
    n = 0;
    return (s1 - (s2 * s3)) / (s4 - (s2 * s2));
}

/*  fitSlopeToTrend  --  Linear regression fit slope to trend for a
             range of days within a month.  Calls the
             more general routines above which are capable
             of spanning month boundaries.  */

double fitSlopeToTrend(struct hdMonth *m, int firstDay, int lastDay)
{
    int i, n;
    
    assert(firstDay <= lastDay);
    assert(lastDay < ((int) m->days));
    
    n = (lastDay - firstDay) + 1;
    
    fitSlopeStart();
    
    for (i = 0; i < n; i++) {
    fitSlopePoint(cvtTrend(m->trendArray[firstDay + i], (weightunit) m->weightUnit, (weightunit) m->weightUnit));
    }
    
    return fitSlopeEnd();
}

/*  fitSlopeToMultiMonthTrend  --  Fit trend to days in the given
                   range of dates.  If the start date
                   or end dates are not present in the
                   database, an error is signified by
                   returning false.  Otherwise the slope
                   of the trend is stored into the slope
                   argument, compensating if necessary for
                   missing months between the start and end.
                   It is permissible to call this function
                   with the same year and month for start
                   and end, but if you know the range is
                                   within a single month it's faster to use
                   fitSlopeToTrend above.

                   You must ensure that the months in
                   question are stored in the memory database
                   before calling this function.  We do
                   try to fake data for missing months, but
                   the quality of the fit will be degraded.

                   Individual monthly logs may be in a
                   variety of weight units.  Trend computation
                   will be done after converting them, if
                   necessary, to the weight unit used by
                   the log for the first month.  Note that
                                   the slope doesn't depend on the weight
                   unit, so all that matters is that we
                   use a consistent unit for all months.
*/

Boolean fitSlopeToMultiMonthTrend(int sYear, int sMonth, int sDay,
                  int eYear, int eMonth, int eDay,
                  double *slope)
{
    struct hdMonth *sm = NULL, *em = NULL;
    int i, sday;
    Boolean ok = false;
    trend ltrend;
    weightunit wu;
    
    sm = mdFindMonth(sYear, sMonth);
    if ((sYear == eYear) && (sMonth == eMonth)) {
    em = sm;
    } else {
    em = mdFindMonth(eYear, eMonth);
    }
    
    if ((sm != NULL) && (em != NULL)) {
    wu = (weightunit) sm->weightUnit;
    fitSlopeStart();
    sday = sDay;
    while (sm != em) {
        if (sm != NULL) {
        for (i = sday; i < ((int) sm->days); i++) {
            fitSlopePoint(cvtTrend(sm->trendArray[i], (weightunit) sm->weightUnit, wu));
        }
        ltrend = cvtTrend(sm->trendArray[sm->days - 1], (weightunit) sm->weightUnit, wu);
        } else {
        /*  Month within range not in database; fake it.  */
        
        for (i = sday; i < DaysInMonth(sMonth + 1, sYear); i++) {
            fitSlopePoint(ltrend);
        }
        }
        sday = 0;
        sMonth++;
        if (sMonth >= 12) {
        sYear++;
        sMonth = 0;
        }
        if ((sYear == eYear) && (sMonth == eMonth)) {
        sm = em;
        } else {
        sm = mdFindMonth(sYear, sMonth);
        }
    }
    for (i = sday; i <= eDay; i++) {
        fitSlopePoint(cvtTrend(em->trendArray[i], (weightunit) em->weightUnit, wu));
    }
    *slope = fitSlopeEnd();
    ok = true;
    }
    return ok;
}
